/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 */

define('io.ox/office/spreadsheet/view/mixin/headertrackingmixin', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/utils/tracking',
    'io.ox/office/spreadsheet/utils/paneutils'
], function (Utils, Tracking, PaneUtils) {

    'use strict';

    // mix-in class HeaderTrackingMixin =======================================

    /**
     * Mix-in class for the class HeaderPane that implements all kinds of mouse
     * and touch tracking, for example selection, or manipulation of column/row
     * sizes.
     *
     * @constructor
     *
     * @param {String} paneSide
     *  The pane side identifier of this header pane.
     */
    function HeaderTrackingMixin(paneSide) {

        // self reference
        var self = this;

        // whether this header pane contains columns
        var columns = PaneUtils.isColumnSide(paneSide);

        // the spreadsheet model and view
        var docView = this.getDocView();
        var docModel = docView.getDocModel();
        var app = docModel.getApp();

        // the (scrolled) content node of the header pane
        var contentNode = this.getContentNode();

        // the type of the current tracking
        var trackingType = null;

        // whether current tracking can be continued in read-only mode
        var readOnlyTracking = false;

        // private methods ----------------------------------------------------

        /**
         * Handles a changed document edit mode. If switched to read-only mode,
         * cancels in-place cell edit mode, or a running tracking action to
         * move or resize drawing objects.
         */
        function editModeHandler(event, editMode) {
            // stop tracking if edit rights have been lost
            if (!editMode && !readOnlyTracking) {
                Tracking.cancelTracking();
            }
        }

        /**
         * Cancels resize tracking when the selection changes.
         */
        function changeSelectionHandler() {
            if (trackingType === 'resize') {
                Tracking.cancelTracking();
            }
        }

        /**
         * Returns the layout data of the column/row matching the 'data-index'
         * attribute of the target DOM node in the passed 'tracking:start'
         * event.
         *
         * @param {jQuery.Event} event
         *  The 'tracking:start' event.
         *
         * @returns {Object}
         *  The layout data of the matching column/row, in the properties
         *  'index', 'offset', and 'size'.
         */
        function getTrackingStartEntryData(event) {
            var indexNode = $(event.target).closest('[data-index]');
            return self.getColRowCollection().getEntry(Utils.getElementAttributeAsInteger(indexNode, 'data-index', -1));
        }

        /**
         * Returns the current absolute sheet position in the header pane for
         * the passed tracking event.
         *
         * @param {jQuery.Event} event
         *  The tracking event.
         *
         * @returns {Number}
         *  The absolute sheet position, in pixels.
         */
        function getTrackingOffset(event) {
            return (columns ? (event.pageX - contentNode.offset().left) : (event.pageY - contentNode.offset().top)) + self.getRenderPosition().offset;
        }

        /**
         * Returns the column/row collection entry matching the position in the
         * passed tracking event.
         *
         * @param {jQuery.Event} event
         *  The tracking event.
         *
         * @returns {Object}
         *  The column/row collection entry for the tracking position.
         */
        function getTrackingEntry(event, options) {
            return self.getColRowCollection().getEntryByOffset(getTrackingOffset(event), Utils.extendOptions(options, { pixel: true }));
        }

        /**
         * Handles all tracking events while cell selection is active, and
         * triggers the appropriate selection events to the own listeners.
         */
        var selectionTrackingHandler = (function () {

            // current index (prevent updates while tracking over the same column/row)
            var currentIndex = 0;

            // initializes selection tracking
            function initializeTracking(event) {
                currentIndex = getTrackingStartEntryData(event).index;
                self.scrollToEntry(currentIndex);
                docView.setSheetViewAttributes({
                    activePane: PaneUtils.getNextPanePos(docView.getSheetViewAttribute('activePane'), paneSide),
                    activePaneSide: paneSide
                });
                self.trigger('select:start', currentIndex, PaneUtils.getSelectionMode(event));
            }

            // updates the current index according to the passed tracking event
            function updateTracking(event) {
                var index = getTrackingEntry(event, { outerHidden: true }).index;
                if (index !== currentIndex) {
                    currentIndex = index;
                    self.trigger('select:move', currentIndex);
                }
            }

            // finalizes selection tracking
            function finalizeTracking(event, success) {
                self.scrollToEntry(currentIndex);
                docView.setSheetViewAttribute('activePaneSide', null);
                self.trigger('select:end', currentIndex, success);

                // ios fires no native 'contextmenu'-event, so i have to trigger it
                if (_.device('ios') && event.duration > 750) {
                    contentNode.find('.cell[data-index="' + currentIndex + '"]').trigger('contextmenu');
                }
            }

            // return the actual selectionTrackingHandler() method
            return function (event) {
                switch (event.type) {
                    case 'tracking:start':
                        initializeTracking(event);
                        break;
                    case 'tracking:move':
                        updateTracking(event);
                        event.preventDefault(); // prevent native scrolling on touch devices
                        break;
                    case 'tracking:scroll':
                        self.scrollRelative(columns ? event.scrollX : event.scrollY);
                        updateTracking(event);
                        break;
                    case 'tracking:end':
                        finalizeTracking(event, true);
                        break;
                    case 'tracking:cancel':
                        finalizeTracking(event, false);
                        break;
                }
            };
        }()); // end of selectionTrackingHandler() local scope

        /**
         * Handles all tracking events while resizing a column/row is active.
         */
        var resizeTrackingHandler = (function () {

            // the layout data of the resized column/row
            var entryData = null;
            // difference between exact start position and trailing position of column/row
            var correction = 0;
            // the current size of the column/row while tracking, in pixels
            var currentSize = 0;
            // minimum/maximum scroll position allowed while resizing
            var minScrollPos = 0, maxScrollPos = 0;
            // maximum size of columns/rows
            var maxSize = Utils.convertHmmToLength(docModel.getMaxColRowSizeHmm(columns), 'px', 1);

            // initializes column/row resizing according to the passed tracking event
            function initializeTracking(event) {

                var position = self.getVisiblePosition();
                var hiddenSize = position.offset - self.getScrollPos();
                var resizeEvent = new $.Event('resize:start');

                entryData = getTrackingStartEntryData(event);
                // calculate difference between exact tracking start position and end of column/row
                // (to prevent unwanted resize actions without moving the tracking point)
                correction = getTrackingOffset(event) - (entryData.offset + entryData.size);
                currentSize = entryData.size;
                minScrollPos = Math.max(0, entryData.offset - 20 - hiddenSize);
                maxScrollPos = Math.max(minScrollPos, entryData.offset + maxSize + 20 - position.size - hiddenSize);
                self.trigger(resizeEvent, entryData.offset, currentSize);
                if (resizeEvent.isDefaultPrevented()) { Tracking.cancelTracking(); }
            }

            // updates column/row resizing according to the passed tracking event
            function updateTracking(event) {
                currentSize = Utils.minMax(getTrackingOffset(event) - correction - entryData.offset, 0, maxSize);
                self.trigger('resize:move', entryData.offset, currentSize);
            }

            // updates scroll position according to the passed tracking event
            function updateScroll(event) {

                var scrollPos = self.getScrollPos();
                var newScrollPos = scrollPos + (columns ? event.scrollX : event.scrollY);

                if ((newScrollPos < scrollPos) && (scrollPos > minScrollPos)) {
                    self.scrollTo(Math.max(newScrollPos, minScrollPos));
                } else if ((newScrollPos > scrollPos) && (scrollPos < maxScrollPos)) {
                    self.scrollTo(Math.min(newScrollPos, maxScrollPos));
                }
            }

            // finalizes the resize tracking
            function finalizeTracking() {
                self.trigger('resize:end');
            }

            // execute the resize operation
            function applyResize() {
                if (currentSize !== entryData.size) {
                    // pass tracked column/row as target index, will cause to modify all columns/rows selected completely
                    docView[columns ? 'setColumnWidth' : 'setRowHeight'](docView.convertPixelToHmm(currentSize), { custom: true, target: entryData.index });
                }
            }

            // return the actual resizeTrackingHandler() method
            return function (event) {
                switch (event.type) {
                    case 'tracking:start':
                        initializeTracking(event);
                        event.preventDefault(); // prevent native scrolling on touch devices
                        break;
                    case 'tracking:move':
                        updateTracking(event);
                        event.preventDefault(); // prevent native scrolling on touch devices
                        break;
                    case 'tracking:scroll':
                        updateScroll(event);
                        updateTracking(event);
                        break;
                    case 'tracking:end':
                        finalizeTracking(event);
                        applyResize();
                        break;
                    case 'tracking:cancel':
                        finalizeTracking(event);
                        break;
                }
            };
        }()); // end of resizeTrackingHandler() local scope

        /**
         * Handles 'tracking:start' events end decides whether to select cells,
         * or to resize columns or rows.
         */
        function trackingStartHandler(event) {

            // target node
            var targetNode = $(event.target);

            // starts tracking with the specified type and event handler
            function startTracking(type, trackingHandler, options) {

                // store the tracking type
                trackingType = type;
                // store whether tracking in read-only mode is supported
                readOnlyTracking = Utils.getBooleanOption(options, 'readOnlyTracking', false);

                // handle all tracking events for this tracking cycle
                PaneUtils.processTrackingCycle(event, trackingHandler, function () {
                    trackingType = null;
                    if (!Utils.TOUCH) {
                        docView.grabFocus();
                    }
                });
            }

            if (targetNode.closest('.cell').length > 0) {
                startTracking('select', selectionTrackingHandler, { readOnlyTracking: true });
            } else if (targetNode.closest('.resizer,.touchresizer').length > 0) {
                startTracking('resize', resizeTrackingHandler);
            }
        }

        /**
         * Handles 'dblclick' events on resizer nodes and resizes the
         * column/row to the optimal width/height depending on cell content.
         */
        function doubleClickHandler(event) {

            // the column/row index
            var index = self.getColRowCollection().getEntry(Utils.getElementAttributeAsInteger(event.target, 'data-index', -1)).index;

            if (columns) {
                docView.setOptimalColumnWidth(index);
            } else {
                docView.setOptimalRowHeight(index);
            }

            // return focus to active grid pane, after the event has been processed
            self.executeDelayed(function () { docView.grabFocus(); }, { infoString: 'HeaderTrackingMixin: doubleClickHandler', app: app });
        }

        // initialization -----------------------------------------------------

        // enable/disable operations that modify the document according to edit mode
        this.listenTo(docModel, 'change:editmode', editModeHandler);

        // listen to view events
        this.listenTo(docView, 'change:selection', changeSelectionHandler);

        // tracking for column/row selection with mouse/touch
        this.getNode().on('tracking:start', trackingStartHandler);

        // double click on resizer for column/row
        this.getNode().on('dblclick', '.resizer', doubleClickHandler);

        // destroy all class members on destruction
        this.registerDestructor(function () {
            self = docModel = docView = contentNode = null;
        });

    } // class HeaderTrackingMixin

    // exports ================================================================

    return HeaderTrackingMixin;

});
