/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Ingo Schmidt-Rosbiegal <ingo.schmidt-rosbiegal@open-xchange.com>
 */

define('io.ox/office/textframework/format/tablestyles', [
    'io.ox/office/tk/utils',
    'io.ox/office/editframework/utils/color',
    'io.ox/office/editframework/utils/border',
    'io.ox/office/editframework/utils/lineheight',
    'io.ox/office/editframework/utils/attributeutils',
    'io.ox/office/editframework/model/stylecollection',
    'io.ox/office/textframework/utils/dom',
    'io.ox/office/textframework/components/table/table',
    'io.ox/office/editframework/utils/gradient'

], function (Utils, Color, Border, LineHeight, AttributeUtils, StyleCollection, DOM, Table, Gradient) {

    'use strict';

    var // definitions for table attributes
        DEFINITIONS = {

            /**
             * Width of the table, as number in 1/100 of millimeters.
             */
            width: {
                def: 'auto',
                format: function (element, width) {
                    var parentWidth = 0;
                    // TODO: Adding support for width in percent
                    // -> better for changes of page size or page orientation
                    if (width === 'auto') {
                        element.css('width', '100%');
                    } else {
                        // adapt size of table to its parent (36132,36163)
                        width = Utils.convertHmmToLength(width, 'px', 1);
                        parentWidth = $(element).parent().width();
                        element.css('width', (parentWidth < width) ? '100%' : (width + 'px'));
                    }
                }
            },

            /**
             * Fill color of the table.
             */
            fillColor: {
                def: Color.AUTO,
                format: function (element, color) {
                    element.css('background-color', this.getCssColor(color, 'fill'));
                },
                preview: function (element, color) {
                    var colorDesc = this.resolveColor(Color.parseJSON(color), 'fill');
                    if (colorDesc.a > 0) {
                        element.css('background-color', colorDesc.css);
                    }
                }
            },

            /**
             * Grid width of columns in relative units. It is an array of numbers
             */
            tableGrid: {
                def: [],
                scope: 'element'
            },

            /**
             * Array containing information, if conditional attributes will be
             * used. As default value, all styles will be used, so that this
             * array can be empty.
             */
            exclude: {
                def: [],
                scope: 'element'
            },

            /**
             * Left border of the table (will be set in the table cells).
             */
            borderLeft: { def: Border.NONE },

            /**
             * Top border of the table (will be set in the table cells).
             */
            borderTop: { def: Border.NONE },

            /**
             * Right border of the table (will be set in the table cells).
             */
            borderRight: { def: Border.NONE },

            /**
             * Bottom border of the table (will be set in the table cells).
             */
            borderBottom: { def: Border.NONE },

            /**
             * Inner horizontal borders inside the table (will be set in the
             * table cells).
             */
            borderInsideHor: { def: Border.NONE },

            /**
             * Inner vertical borders inside the table (will be set in the
             * table cells).
             */
            borderInsideVert: { def: Border.NONE },

            /**
             * Top padding of the whole table
             */
            paddingTop: { def: 0 },

            /**
             * Bottom padding of the whole table
             */
            paddingBottom: { def: 0 },

            /**
             * Left padding of the whole table
             */
            paddingLeft: { def: 0 },

            /**
             * Right padding of the whole table
             */
            paddingRight: { def: 0 }

        },

        // the conditional table attributes (only 'wholeTable' has no geometric dependencies)
        GEOMETRIC_TABLE_ATTRIBUTES = ['firstRow', 'lastRow', 'firstCol', 'lastCol', 'band1Vert', 'band2Vert', 'band1Hor', 'band2Hor', 'northEastCell', 'northWestCell', 'southEastCell', 'southWestCell'],

        // the border properties
        BORDER_PROPS = ['borderTop', 'borderBottom', 'borderLeft', 'borderRight', 'borderInsideHor', 'borderInsideVert'];

    // class TableStyles ======================================================

    /**
     * Contains the style sheets for table formatting attributes. The CSS
     * formatting will be written to table elements and their rows and cells.
     *
     * @constructor
     *
     * @extends StyleCollection
     *
     * @param {TextModel} docModel
     *  The text document model containing instance.
     */
    function TableStyles(docModel) {

        var // self reference
            self = this,
            app = docModel.getApp();

        // base constructor ---------------------------------------------------

        StyleCollection.call(this, docModel, 'table', {
            families: 'cell paragraph character changes',
            styleAttributesResolver: resolveTableStyleAttributes,
            elementAttributesResolver: resolveTableElementAttributes,
            formatHandler: updateTableFormatting,
            previewHandler: updateTablePreviewFormatting
        });

        // private methods ----------------------------------------------------

        /**
         * Returns the attributes of the specified attribute family contained
         * in table style sheets. Resolves the conditional attributes that
         * match the position of the passed source element.
         *
         * @param {Object} styleAttributes
         *  The complete 'attributes' object of a table style sheet.
         *
         * @param {jQuery} tableNode
         *  The DOM table node referring to a table style sheet, as jQuery
         *  object.
         *
         * @param {jQuery} [cellNode]
         *  The DOM cell node that has initially requested the formatting
         *  attributes of a table style sheet, as jQuery object.
         *
         * @returns {Object}
         *  The formatting attributes extracted from the passed style sheet
         *  attributes object, as map of attribute value maps (name/value
         *  pairs), keyed by attribute family.
         */
        function resolveTableStyleAttributes(styleAttributes, tableNode, cellNode) {

            var // table size
                lastTableRow = Table.getRowCount(tableNode) - 1,
                lastTableCol = Table.getColumnCount(tableNode) - 1,

                // the grid row range covered by the cell
                rowRange = (cellNode.length > 0) ? Table.getGridRowRangeOfCell(cellNode) : null,
                isFirstRow = rowRange && (rowRange.start === 0),
                isLastRow = rowRange && (rowRange.end === lastTableRow),

                // the grid grid column range covered by the cell
                colRange = (cellNode.length > 0) ? Table.getGridColumnRangeOfCell(cellNode) : null,
                isFirstCol = colRange && (colRange.start === 0),
                isLastCol = colRange && (colRange.end === lastTableCol),

                // the excluded conditional keys
                excludedConditionalKeys = AttributeUtils.getExplicitAttributes(tableNode, { family: 'table', direct: true }).exclude,
                firstRowIncluded = isConditionalKeyIncluded('firstRow'),
                lastRowIncluded = isConditionalKeyIncluded('lastRow'),
                firstColIncluded = isConditionalKeyIncluded('firstCol'),
                lastColIncluded = isConditionalKeyIncluded('lastCol'),

                // conditional key of the inner horizontal/vertical bands, according to cell position
                bandKey = null, bandSize = 1, bandIndex = 0,

                // the extracted style attributes according to the position of the table cell
                attributes = {};

            // whether the passed conditional key is included according to the 'exclude' attribute
            function isConditionalKeyIncluded(conditionalKey) {
                return !_.isArray(excludedConditionalKeys) || !_(excludedConditionalKeys).contains(conditionalKey);
            }

            // copies a border attribute (table or cell) to an outer cell border attribute
            function updateOuterCellBorder(sourceAttributes, outerBorderName, innerBorderName, isOuterBorder) {

                var // table and cell attributes (either may be missing)
                    tableAttributes = _.isObject(sourceAttributes.table) ? sourceAttributes.table : {},
                    cellAttributes = _.isObject(sourceAttributes.cell) ? sourceAttributes.cell : {},
                    // the source border attribute value
                    border = null;

                if (isOuterBorder) {
                    // copy outer table border to cell border if border is missing in cell attributes
                    border = _.isObject(cellAttributes[outerBorderName]) ? null : tableAttributes[outerBorderName];
                } else {
                    // copy inner table or cell border (cell border wins) to outer cell border
                    border = _.isObject(cellAttributes[innerBorderName]) ? cellAttributes[innerBorderName] : tableAttributes[innerBorderName];
                }

                // insert existing border to the specified outer cell border
                if (_.isObject(border)) {
                    (attributes.cell || (attributes.cell = {}))[outerBorderName] = border;
                }
            }

            // copies a padding attribute from a table to a cell, if required
            function updateCellPadding(sourceAttributes, paddingName) {

                var // table and cell attributes (either may be missing)
                    tableAttributes = _.isObject(sourceAttributes.table) ? sourceAttributes.table : {},
                    cellAttributes = _.isObject(sourceAttributes.cell) ? sourceAttributes.cell : {},
                    // the padding attribute value
                    padding = (paddingName in cellAttributes) ? cellAttributes[paddingName] : tableAttributes[paddingName];

                // insert existing border to the specified outer cell border
                if (_.isNumber(padding)) {
                    (attributes.cell || (attributes.cell = {}))[paddingName] = padding;
                }
            }

            // merges the specified conditional attributes into the 'attributes' object
            function mergeConditionalAttributes(conditionalKey, isHorizontalBand, isVerticalBand) {

                var // the attributes at the passed conditional key
                    conditionalAttributes = styleAttributes[conditionalKey];

                if (_.isObject(conditionalAttributes)) {

                    // copy all attributes from the style sheet to the result object
                    docModel.extendAttributes(attributes, conditionalAttributes);

                    // INFO: In Presentation app it is possible, that 'borderInsideHor' and 'borderInsideVert' are set
                    //       to the 'cell' family in table styles. Therefore it can happen, that these properties are
                    //       set via the conditionalAttributes although this might not be wanted (48609). Then this
                    //       properties need to be overwritten with hard attributes in operations.

                    // copy inner borders to outer cell borders, if cell is located inside the current table area
                    updateOuterCellBorder(conditionalAttributes, 'borderTop', 'borderInsideHor', !isVerticalBand || isFirstRow);
                    updateOuterCellBorder(conditionalAttributes, 'borderBottom', 'borderInsideHor', !isVerticalBand || isLastRow);
                    updateOuterCellBorder(conditionalAttributes, 'borderLeft', 'borderInsideVert', !isHorizontalBand || isFirstCol);
                    updateOuterCellBorder(conditionalAttributes, 'borderRight', 'borderInsideVert', !isHorizontalBand || isLastCol);

                    updateCellPadding(conditionalAttributes, 'paddingTop');
                    updateCellPadding(conditionalAttributes, 'paddingBottom');
                    updateCellPadding(conditionalAttributes, 'paddingLeft');
                    updateCellPadding(conditionalAttributes, 'paddingRight');
                }
            }

            // in presentation app the border attributes are assigned to 'cell' family, not 'table' family. But for correct
            // resolving of table border properties it is necessary, that they are assigned to the 'table' family, too (48647).
            function transferWholeTableBordersToTable() {
                if (styleAttributes.wholeTable.cell) {
                    _.each(BORDER_PROPS, function (prop) {
                        if (styleAttributes.wholeTable.cell[prop] && (!styleAttributes.wholeTable.table || !styleAttributes.wholeTable.table[prop])) {
                            if (!attributes.table) { attributes.table = {}; }
                            attributes.table[prop] = _.copy(styleAttributes.wholeTable.cell[prop], true);
                        }
                    });
                }
            }

            // in presentation app the border attributes are set at the 'cell' family
            if (docModel.useSlideMode() && styleAttributes.wholeTable) { transferWholeTableBordersToTable(); }

            // wholeTable: always
            mergeConditionalAttributes('wholeTable', true, true);

            // inner horizontal bands
            if (rowRange && !(firstRowIncluded && isFirstRow) && !(lastRowIncluded && isLastRow) && isConditionalKeyIncluded('bandsHor')) {
                // size of horizontal bands, TODO: replace '1' with attribute value
                bandSize = Math.floor(Math.max(1, 1));
                // ignore first row in calculation of band index
                bandIndex = Math.floor((rowRange.start - (firstRowIncluded ? 1 : 0)) / bandSize);
                // resolve odd or even band attributes
                bandKey = ((bandIndex % 2) === 0) ? 'band1Hor' : 'band2Hor';
                mergeConditionalAttributes(bandKey, true, false);
            }

            // inner vertical bands
            if (colRange && !(firstColIncluded && isFirstCol) && !(lastColIncluded && isLastCol) && isConditionalKeyIncluded('bandsVert')) {
                // size of vertical bands, TODO: replace '1' with attribute value
                bandSize = Math.floor(Math.max(1, 1));
                // ignore first column in calculation of band index
                bandIndex = Math.floor((colRange.start - (firstColIncluded ? 1 : 0)) / bandSize);
                // resolve odd or even band attributes
                bandKey = ((bandIndex % 2) === 0) ? 'band1Vert' : 'band2Vert';
                mergeConditionalAttributes(bandKey, false, true);
            }

            // cell inside first/last row/column
            if (firstColIncluded && isFirstCol) {
                mergeConditionalAttributes('firstCol', false, true);
            }
            if (lastColIncluded && isLastCol) {
                mergeConditionalAttributes('lastCol', false, true);
            }
            if (firstRowIncluded && isFirstRow) {
                mergeConditionalAttributes('firstRow', true, false);
            }
            if (lastRowIncluded && isLastRow) {
                mergeConditionalAttributes('lastRow', true, false);
            }

            // single corner cells (only if inside active first/last row AND column areas)
            if (firstRowIncluded && firstColIncluded && isFirstRow && isFirstCol) {
                mergeConditionalAttributes('northWestCell', false, false);
            }
            if (firstRowIncluded && lastColIncluded && isFirstRow && isLastCol) {
                mergeConditionalAttributes('northEastCell', false, false);
            }
            if (lastRowIncluded && firstColIncluded && isLastRow && isFirstCol) {
                mergeConditionalAttributes('southWestCell', false, false);
            }
            if (lastRowIncluded && lastColIncluded && isLastRow && isLastCol) {
                mergeConditionalAttributes('southEastCell', false, false);
            }

            return attributes;
        }

        /**
         * Returns the attributes of the specified attribute family contained
         * in table style sheets. Resolves the conditional attributes that
         * match the position of the passed source element.
         *
         * @param {Object} elementAttributes
         *  The explicit attributes of the table element, as map of attribute
         *  value maps (name/value pairs), keyed by attribute family.
         *
         * @param {jQuery} tableNode
         *  The DOM table node, as jQuery object.
         *
         * @param {jQuery} [cellNode]
         *  The DOM cell node that has initially requested the formatting
         *  attributes, as jQuery object.
         *
         * @returns {Object}
         *  The resolved explicit formatting attributes, as map of attribute
         *  value maps (name/value pairs), keyed by attribute family.
         */
        function resolveTableElementAttributes(elementAttributes, tableNode, cellNode) {
            return resolveTableStyleAttributes({ wholeTable: elementAttributes }, tableNode, cellNode);
        }

        /**
         * Setting the table background. This can currently only be done using a style.
         * Only supported in presentation app.
         *
         * @param {jQuery} tableNode
         *  The DOM table node, as jQuery object.
         *
         * @param {Object} styleSheetAttributeMap
         *  The style sheet attributes of the table
         */
        function handleTableBackground(table, styleSheetAttributeMap) {

            if (styleSheetAttributeMap && styleSheetAttributeMap.backgroundStyle && styleSheetAttributeMap.backgroundStyle.fill) {

                var // the fill attributes of the slide
                    fillAttrs = styleSheetAttributeMap.backgroundStyle.fill,
                    // the CSS properties for the slide
                    cssProps = {};

                // calculate the CSS fill attributes
                switch (fillAttrs.type) {
                    case 'none':
                        // clear everything: color and bitmaps
                        cssProps.background = '';
                        break;

                    case 'solid':
                        cssProps.background = docModel.getCssColor(fillAttrs.color, 'fill', docModel.getTargetChain(docModel.getSlideIdForNode(table)));
                        break;

                    case 'gradient':
                        var
                            elmSlide  = table[0],
                            target    = docModel.getTargetChain(docModel.getSlideIdForNode(table)),
                            gradient  = Gradient.create(Gradient.getDescriptorFromAttributes(fillAttrs)),
                            dataUrl   = Gradient.parseImageDataUrl(gradient, docModel, { width: elmSlide.offsetWidth, height: elmSlide.offsetHeight }, target);

                        cssProps.background = ['url("', dataUrl, '")'].join('');
                        cssProps.backgroundSize = '100%'; // #49272
                        break;

                    default:
                        Utils.warn('TableStyles.handleTableBackground(): unknown fill type "' + fillAttrs.type + '"');
                }

                // apply the fill attributes
                table.css(cssProps);

            } else {
                table.css({ background: '' }); // removing background
            }

        }

        /**
         * Will be called for every table element whose attributes have been
         * changed. Repositions and reformats the table according to the passed
         * attributes.
         *
         * @param {jQuery} table
         *  The <table> element whose table attributes have been changed, as
         *  jQuery object.
         *
         * @param {Object} mergedAttributes
         *  A map of attribute value maps (name/value pairs), keyed by
         *  attribute family, containing the effective attribute values merged
         *  from style sheets and explicit attributes.
         *
         * @param {Boolean} async
         *  If set to true, the table formatting will be updated asynchronously
         *  in a browser timeout loop.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved when the
         *  the table is completely formatted.
         */
        function updateTableFormatting(table, mergedAttributes, async) {

            var // the table styles/formatter
                tableStyles = docModel.getStyleCollection('table'),
                // the table row styles/formatter
                tableRowStyles = docModel.getStyleCollection('row'),
                // the table cell styles/formatter
                tableCellStyles = docModel.getStyleCollection('cell'),
                // the paragraph styles/formatter
                paragraphStyles = docModel.getStyleCollection('paragraph'),
                // the column widths
                tableGrid = mergedAttributes.table.tableGrid,
                // the explicit table attributes
                tableAttributes = AttributeUtils.getExplicitAttributes(table, { family: 'table' }),
                // all row nodes in the table
                rowNodes = DOM.getTableRows(table),
                // a collector for all cells in the first row
                allCellsInFirstRow = rowNodes.first().children('td').children('div.cell'),
                // the upper left cell of the table (the div.cell element)
                firstCell = allCellsInFirstRow.first(),
                // a counter for the cells (performance)
                cellCounter = allCellsInFirstRow.length * rowNodes.length,
                // whether this update was triggered by a resize of table (column width)
                isTableResize = false,
                // whether the table has a 'simple' style
                isSimpleTableStyle = false,
                // whether cells were removed from the table
                isGUIRemoveOperation = false,
                // the number of the first row, that requires a formatting update
                minUpdateRow = null,
                // the style sheet attributes for the table
                styleSheetAttributeMap = null,
                // the result Deferred object
                def = null;

            // checking if the attributes have table geometric dependencies
            // -> then a complete reformatting of the table is required.
            // Otherwise only the new cells (insert operations) or no cells
            // (delete operations) need to be formatted.
            function checkSimpleTableStyle() {

                var isSimple = true;

                _.each(GEOMETRIC_TABLE_ATTRIBUTES, function (attribute) {
                    if ((attribute in styleSheetAttributeMap) && !_.contains(tableAttributes.exclude, attribute)) {
                        isSimple = false;
                    }
                });
                return isSimple;
            }

            // helper function to update the top and the left resizers in the table cell
            function updateTopLeftResizers() {

                // the old upper left cell is no longer the current upper left cell
                table.find('.upperLeftCell').removeClass('upperLeftCell').removeData('cellcount');

                // removing old top and left resizer nodes
                table.find('.resize.top, .resize.left').remove();

                // adding new top and left resizers to first row and first column
                allCellsInFirstRow.prepend($('<div>').addClass('resize top'));

                rowNodes.children('td:first-child').children('div.cell').prepend($('<div>').addClass('resize left'));

                // marking the new upper left cell (saving cell count for performance reasons)
                firstCell.addClass('upperLeftCell').data('cellcount', cellCounter);
            }

            // updates the formatting of the passed table row
            function updateRowFormatting(rowNode) {

                // update the table row itself
                tableRowStyles.updateElementFormatting(rowNode);

                // update table cells (do NOT pass merged attributes passed
                // to the updateTableFormatting() method as base attributes,
                // attributes must be recalculated for each single cell!)
                $(rowNode).children('td').each(function () {

                    var cellFormatting = true;

                    // Performance: Updating only new cells, if this update was triggered from a GUI event
                    // and if the table has a simple table style -> otherwise all cells need to be updated
                    // Performance: Simplified cell formatting for deleteColumn and deleteRow GUI operations.
                    // For the insert operations (insertRow or insertColumn) cell (and paragraph) formatting
                    // is always required.
                    if (isGUIRemoveOperation) {
                        if (isSimpleTableStyle) {
                            cellFormatting = false;  // using simplified cell formatting update (only taking care of tab stops)
                            // TODO: For not simple table styles the full cell formatting is only required for those cells
                            // that are behind the new/removed cell in the row for insertColumn and deleteColumn
                        }
                    }

                    // Performance: Resizing the table (column width) requires only tab stop recalculation
                    if (isTableResize) { cellFormatting = false; }

                    if (cellFormatting) {
                        // complete formatting update of the cell -> this is necessary for complex table styles, but also for simple
                        // table styles, because paragraph formatting is also necessary. For example determines the paragraph height the
                        // height of the row, if the table row height is set to 'auto'.
                        tableCellStyles.updateElementFormatting(this);
                    } else {
                        // checking if a cell contains tab stops, at least these need to be updated, if cell width was changed
                        if ($(this).find(DOM.TAB_NODE_SELECTOR).length > 0) {
                            // updating the tabs is always required, because the width of a cell might have changed -> updating all paragraphs in the table cell
                            Utils.iterateSelectedDescendantNodes(DOM.getCellContentNode(this), DOM.PARAGRAPH_NODE_SELECTOR, function (paragraph) {
                                paragraphStyles.updateTabStops(paragraph);
                            }, undefined, { children: true });
                        }
                    }
                });
            }

            // checking if the attributes have table geometric dependencies or table background
            if (mergedAttributes.styleId) {
                styleSheetAttributeMap = tableStyles.getStyleSheetAttributeMap(mergedAttributes.styleId);
                isSimpleTableStyle = checkSimpleTableStyle();
            }

            // handling of table background
            // if (docModel.useSlideMode()) { handleTableBackground(table, styleSheetAttributeMap); }

            if (table.data('gui')) {
                if (table.data('gui') === 'remove') {
                    isGUIRemoveOperation = true;
                }
                table.removeData('gui');
            }

            if (table.data('tableResize')) {
                isTableResize = true;
                table.removeData('tableResize');
            }

            // reformatting table colgroup in Firefox during pasting internal clip board, task 29401, task 30477
            if ((table.data('internalClipboard') || table.data('undoRedoRunning')) && _.browser.Firefox) {
                table.removeData('internalClipboard');
                table.removeData('undoRedoRunning');
            }

            // update column widths (also required for bug fixes 29212 and 29675)
            Table.updateColGroup(table, tableGrid);

            // Fix for 30821 (on webkit browsers), max-width of 100% is ignored, if width is defined
            if ((_.browser.WebKit) && (table.closest('div.cellcontent').length > 0) && (table.css('max-width') === '100%') && (table.width() > table.closest('div.cellcontent').width())) {
                table.css('width', '');
            }

            // Performance: Immediately return, if no update required for simple tables (for example after deleteRows).
            // Even update of tab stops is not required for insertRow and deleteRow. But insertRow requires update of
            // paragraph attributes.
            if (_.isNumber(table.data('reducedTableFormatting'))) {
                minUpdateRow = table.data('reducedTableFormatting');
                table.removeData('reducedTableFormatting');
                if ((isSimpleTableStyle) && (isGUIRemoveOperation)) {
                    return $.when();
                }
            }

            // Performance check for very slow devices. In this case an asynchronous update is
            // required, so that the user gets faster feed back
            async = async || (Utils.IOS && Utils.SLOW_DEVICE);

            // updating the top and left resize markers in the table. This is necessary, if the
            // upper left cell is no longer the upper left cell of the table or if the number of
            // cells in the table has changed (insertRow, ...)
            if (docModel.useSlideMode() && firstCell.length > 0 && (!firstCell.hasClass('upperLeftCell') || firstCell.data('cellcount') !== cellCounter)) { updateTopLeftResizers(); }

            // Performance: Reducing the number of rows, because upper rows require no update
            if (_.isNumber(minUpdateRow)) {
                rowNodes = rowNodes.slice(minUpdateRow);
            }

            // change track attribute handling
            docModel.getChangeTrack().updateChangeTrackAttributes(table, mergedAttributes);

            // update rows, cells, ... , but only if required
            if (async) {
                // process the table asynchronously
                def = self.iterateArraySliced(rowNodes, updateRowFormatting, { infoString: 'Text: Table: updateRowFormatting', app: app });
                def.always(function () {
                    if (docModel.useSlideMode()) { handleTableBackground(table, styleSheetAttributeMap); } // (48370)
                    docModel.trigger('table:formatting:done');
                });

            } else {
                // update all rows at once (but asynchronous, so that for example a new column is visible immediately)
                def = self.executeDelayed(function () {
                    rowNodes.each(function () { updateRowFormatting(this); });
                    if (docModel.useSlideMode()) { handleTableBackground(table, styleSheetAttributeMap); } // (48370)
                    docModel.trigger('table:formatting:done');
                }, { infoString: 'TableStyles: updateTableFormatting', app: app });
            }

            // handling table background (48370)
            // def.done(function () { if (docModel.useSlideMode()) { handleTableBackground(table, styleSheetAttributeMap); } });

            return def.promise();
        }

        /**
         * Will be called for tables used as preview elements in the GUI.
         *
         * @param {jQuery} table
         *  The preview table node, as jQuery object.
         *
         * @param {Object} mergedAttributes
         *  A map of attribute maps (name/value pairs), keyed by attribute
         *  family, containing the effective attribute values merged from style
         *  sheets and explicit attributes.
         */
        function updateTablePreviewFormatting(table, mergedAttributes) {

            var // the table cell styles/formatter
                tableCellStyles = docModel.getStyleCollection('cell'),
                // the table styles/formatter
                tableStyles = docModel.getStyleCollection('table'),
                // the the style sheet attributes for the table
                styleSheetAttributeMap = null;

            // setting table backgrounds (48370)
            if (docModel.useSlideMode() && mergedAttributes.styleId) {
                styleSheetAttributeMap = tableStyles.getStyleSheetAttributeMap(mergedAttributes.styleId);
                handleTableBackground(table, styleSheetAttributeMap);
            }

            // iterate over all cells in the table to set the cell attributes
            DOM.getTableCells(table).each(function () {
                // do NOT pass merged attributes passed to this method as base
                // attributes; attributes must be recalculated for each single cell
                tableCellStyles.updateElementFormatting(this, { preview: true });
            });
        }

        // public methods -----------------------------------------------------

        /**
         * Performance: Helper function to update an optional table background.
         * This is only supported in the presentation app.
         *
         * @param {Node|jQuery} drawing
         *  The table drawing node.
         */
        this.updateTableDrawingBackground = function (drawing) {

            // the explicit drawing attributes
            var expAttrs = AttributeUtils.getExplicitAttributes(drawing);
            // the table node inside the drawing
            var tableNode = $(drawing).find(DOM.TABLE_NODE_SELECTOR);

            if (tableNode.length > 0 && expAttrs && expAttrs.styleId) {
                var styleSheetAttributeMap = self.getStyleSheetAttributeMap(expAttrs.styleId);
                handleTableBackground(tableNode, styleSheetAttributeMap);
            }
        };

        // initialization -----------------------------------------------------

        // register the attribute definitions for the style family
        docModel.registerAttributeDefinitions('table', DEFINITIONS);

        // destroy all class members on destruction
        this.registerDestructor(function () {
            docModel = self = null;
        });

    } // class TableStyles

    // static methods ---------------------------------------------------------

    TableStyles.getBorderStyleFromAttributes = function (attributes) {

        var allWidths = [],
            width = 'none';

        if ((attributes) && (attributes.borderLeft) && (Border.isVisibleBorder(attributes.borderLeft))) { allWidths.push(attributes.borderLeft.width); }
        if ((attributes) && (attributes.borderRight) && (Border.isVisibleBorder(attributes.borderRight))) { allWidths.push(attributes.borderRight.width); }
        if ((attributes) && (attributes.borderTop) && (Border.isVisibleBorder(attributes.borderTop))) { allWidths.push(attributes.borderTop.width); }
        if ((attributes) && (attributes.borderBottom) && (Border.isVisibleBorder(attributes.borderBottom))) { allWidths.push(attributes.borderBottom.width); }
        if ((attributes) && (attributes.borderInsideHor) && (Border.isVisibleBorder(attributes.borderInsideHor))) { allWidths.push(attributes.borderInsideHor.width); }
        if ((attributes) && (attributes.borderInsideVert) && (Border.isVisibleBorder(attributes.borderInsideVert))) { allWidths.push(attributes.borderInsideVert.width); }

        allWidths = _.uniq(allWidths);

        if (allWidths.length === 1) {
            width = allWidths[0];
            width = Utils.convertHmmToLength(width, 'pt', 0.1);  // converting from 1/100 mm to pt
        }

        return width;
    };

    TableStyles.getAttributesFromBorderStyle = function (borderWidth, attributes) {

        var borderLeft = _.clone(attributes.borderLeft),
            borderRight = _.clone(attributes.borderRight),
            borderTop = _.clone(attributes.borderTop),
            borderBottom = _.clone(attributes.borderBottom),
            borderInsideHor = _.clone(attributes.borderInsideHor),
            borderInsideVert = _.clone(attributes.borderInsideVert);

        // converting from pt to 1/100 mm
        borderWidth = Utils.convertLengthToHmm(borderWidth, 'pt');

        borderLeft.width = borderWidth;
        borderRight.width = borderWidth;
        borderTop.width = borderWidth;
        borderBottom.width = borderWidth;
        borderInsideHor.width = borderWidth;
        borderInsideVert.width = borderWidth;

        return {
            borderLeft:       borderLeft,
            borderRight:      borderRight,
            borderTop:        borderTop,
            borderBottom:     borderBottom,
            borderInsideHor:  borderInsideHor,
            borderInsideVert: borderInsideVert
        };

    };

    // Light Shading table style attributes

    TableStyles.getLightShadingTableStyleAttributes = function (colorValue) {

        var SINGLE_BORDER_LINE = { style: 'single', color: { type: 'scheme', value: colorValue }, width: 35 },
            BAND_COLOR = { type: 'scheme', value: colorValue, transformations: [{ type: 'tint', value: 24706 }] },
            ROW_CELL_ATTRIBUTES = { borderLeft: Border.NONE, borderTop: SINGLE_BORDER_LINE, borderRight: Border.NONE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE },
            BAND_CELL_ATTRIBUTES = { borderLeft: Border.NONE, borderRight: Border.NONE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: BAND_COLOR };

        return {
            wholeTable: {
                table: { borderTop: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE },
                paragraph: { lineHeight: LineHeight.SINGLE, marginBottom: 0 },
                character: { color: { type: 'scheme', value: colorValue, transformations: [{ type: 'shade', value: 74902 }] } }
            },
            firstRow: { cell: ROW_CELL_ATTRIBUTES, paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 }, character: { bold: true } },
            lastRow: { cell: ROW_CELL_ATTRIBUTES, paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 }, character: { bold: true } },
            firstCol: { character: { bold: true } },
            lastCol: { character: { bold: true } },
            band1Vert: { cell: BAND_CELL_ATTRIBUTES },
            band1Hor: { cell: BAND_CELL_ATTRIBUTES }
        };
    };

    // Medium Shading 1 table style attributes

    TableStyles.getMediumShading1TableStyleAttributes = function (colorValue) {

        var SCHEME_COLOR = { type: 'scheme', value: colorValue },
            BAND_COLOR = { type: 'scheme', value: colorValue, transformations: [{ type: 'tint', value: 24706 }] },
            SINGLE_BORDER_LINE = { style: 'single', color: SCHEME_COLOR, width: 35 },
            DOUBLE_BORDER_LINE = { style: 'double', color: SCHEME_COLOR, width: 26 };

        return {
            wholeTable: {
                table: { borderLeft: SINGLE_BORDER_LINE, borderTop: SINGLE_BORDER_LINE, borderRight: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: SINGLE_BORDER_LINE },
                paragraph: { lineHeight: LineHeight.SINGLE, marginBottom: 0 }
            },
            firstRow: {
                cell: { borderLeft: SINGLE_BORDER_LINE, borderTop: SINGLE_BORDER_LINE, borderRight: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: SCHEME_COLOR },
                paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 },
                character: { bold: true, color: { type: 'scheme', value: 'background1' } }
            },
            lastRow: {
                cell: { borderLeft: SINGLE_BORDER_LINE, borderTop: DOUBLE_BORDER_LINE, borderRight: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE },
                paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 },
                character: { bold: true }
            },
            firstCol: { character: { bold: true } },
            lastCol: { character: { bold: true } },
            band1Vert: { cell: { fillColor: BAND_COLOR } },
            band1Hor: { cell: { borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: BAND_COLOR } },
            band2Hor: { cell: { borderInsideHor: Border.NONE, borderInsideVert: Border.NONE } }
        };
    };

    // Medium Shading 2 table style attributes

    TableStyles.getMediumShading2TableStyleAttributes = function (colorValue) {

        var SCHEME_COLOR = { type: 'scheme', value: colorValue },
            BACK_COLOR = { type: 'scheme', value: 'background1' },
            BAND_COLOR = { type: 'scheme', value: 'background1', transformations: [{ type: 'shade', value: 84706 }] },
            SINGLE_BORDER_LINE = { style: 'single', color: Color.AUTO, width: 79 },
            DOUBLE_BORDER_LINE = { style: 'double', color: Color.AUTO, width: 26 },
            NORTH_CELL_ATTRIBUTES = { borderLeft: Border.NONE, borderTop: SINGLE_BORDER_LINE, borderRight: Border.NONE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE };

        return {
            wholeTable: {
                table: { borderTop: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE },
                paragraph: { lineHeight: LineHeight.SINGLE, marginBottom: 0 }
            },
            firstRow: {
                cell: { borderLeft: Border.NONE, borderTop: SINGLE_BORDER_LINE, borderRight: Border.NONE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: SCHEME_COLOR },
                paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 },
                character: { bold: true, color: BACK_COLOR }
            },
            lastRow: {
                cell: { borderLeft: Border.NONE, borderTop: DOUBLE_BORDER_LINE, borderRight: Border.NONE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: BACK_COLOR },
                paragraph: { lineHeight: LineHeight.SINGLE, marginTop: 0, marginBottom: 0 },
                character: { color: Color.AUTO }
            },
            firstCol: {
                cell: { borderLeft: Border.NONE, borderTop: Border.NONE, borderRight: Border.NONE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: SCHEME_COLOR },
                character: { bold: true, color: BACK_COLOR }
            },
            lastCol: {
                cell: { borderLeft: Border.NONE, borderRight: Border.NONE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: SCHEME_COLOR },
                character: { bold: true, color: BACK_COLOR }
            },
            band1Vert: { cell: { borderLeft: Border.NONE, borderRight: Border.NONE, borderInsideHor: Border.NONE, borderInsideVert: Border.NONE, fillColor: BAND_COLOR } },
            band1Hor: { cell: { fillColor: BAND_COLOR } },
            northEastCell: { cell: NORTH_CELL_ATTRIBUTES },
            northWestCell: { cell: NORTH_CELL_ATTRIBUTES, character: { color: BACK_COLOR } }
        };
    };

    // Medium Grid 1 table style attributes

    TableStyles.getMediumGrid1TableStyleAttributes = function (colorValue) {

        var BAND_COLOR = { type: 'scheme', value: colorValue, transformations: [{ type: 'tint', value: 49804 }] },
            SINGLE_BORDER_LINE = { style: 'single', color: { type: 'scheme', value: colorValue }, width: 35 },
            THICK_BORDER_LINE = { style: 'single', color: { type: 'scheme', value: colorValue }, width: 79 };

        return {
            wholeTable: {
                table: { borderLeft: SINGLE_BORDER_LINE, borderTop: SINGLE_BORDER_LINE, borderRight: SINGLE_BORDER_LINE, borderBottom: SINGLE_BORDER_LINE, borderInsideHor: SINGLE_BORDER_LINE, borderInsideVert: SINGLE_BORDER_LINE },
                cell: { fillColor: { type: 'scheme', value: colorValue, transformations: [{ type: 'tint', value: 24706 }] } },
                paragraph: { lineHeight: LineHeight.SINGLE, marginBottom: 0 }
            },
            firstRow: { character: { bold: true } },
            lastRow: { cell: { borderTop: THICK_BORDER_LINE }, character: { bold: true } },
            firstCol: { character: { bold: true } },
            lastCol: { character: { bold: true } },
            band1Vert: { cell: { fillColor: BAND_COLOR } },
            band1Hor: { cell: { fillColor: BAND_COLOR } }
        };
    };

    // Light Style 3 table style attributes (Presentation)
    TableStyles.getLightStyle3TableStyleAttributesPresentation = function (colorValue1, colorValue2) {
        return {
            wholeTable: {
                cell: {
                    borderLeft: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderRight: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderInsideHor: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderInsideVert: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } }
                },
                character: { color: { type: 'scheme', value: colorValue2 } }
            },
            band1Hor: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'alpha', value: 20000 }] } } },
            band1Vert: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'alpha', value: 20000 }] } } },
            firstRow: {
                cell: { borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } } },
                character: { bold: true }
            },
            lastRow: {
                cell: { borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } } },
                character: { bold: true }
            }
        };
    };

    // Medium Style 2 table style attributes (Presentation)
    TableStyles.getMediumStyle2TableStyleAttributesPresentation = function (colorValue1, colorValue2, colorValue3) { // 'accent1', 'light1', 'dark1'
        return {
            wholeTable: {
                cell: {
                    fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 20000 }] },
                    borderLeft: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } },
                    borderRight: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } },
                    borderInsideHor: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } },
                    borderInsideVert: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } }
                },
                character: { color: { type: 'scheme', value: colorValue3 } }
            },
            band1Hor: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 40000 }] } } },
            band1Vert: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 40000 }] } } },
            firstCol: {
                cell: { fillColor: { type: 'scheme', value: colorValue1 } },
                character: { color: { type: 'scheme', value: colorValue2 }, bold: true }
            },
            firstRow: {
                cell: {
                    fillColor: { type: 'scheme', value: colorValue1 },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } }
                },
                character: {
                    color: { type: 'scheme', value: colorValue2 },
                    bold: true
                }
            },
            lastCol: {
                cell: { fillColor: { type: 'scheme', value: colorValue1 } },
                character: { color: { type: 'scheme', value: colorValue2 }, bold: true }
            },
            lastRow: {
                cell: {
                    fillColor: { type: 'scheme', value: colorValue1 },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue2 } }
                },
                character: {
                    color: { type: 'scheme', value: colorValue2 },
                    bold: true
                }
            }
        };
    };

    // Medium Style 3 table style attributes (Presentation)
    TableStyles.getMediumStyle3TableStyleAttributesPresentation = function (colorValue1, colorValue2, colorValue3) { // 'accent1', 'light1', 'dark1'
        return {
            wholeTable: {
                cell: {
                    fillColor: { type: 'scheme', value: colorValue2 },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue3 } },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue3 } }
                },
                character: { color: { type: 'scheme', value: colorValue3 } }
            },
            band1Hor: { cell: { fillColor: { type: 'scheme', value: colorValue3, transformations: [{ type: 'tint', value: 20000 }] } } },
            band1Vert: { cell: { fillColor: { type: 'scheme', value: colorValue3, transformations: [{ type: 'tint', value: 20000 }] } } },
            firstCol: {
                cell: { fillColor: { type: 'scheme', value: colorValue1 } },
                character: { color: { type: 'scheme', value: colorValue2 }, bold: true }
            },
            firstRow: {
                cell: { fillColor: { type: 'scheme', value: colorValue1 }, borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue3 } } },
                character: { color: { type: 'scheme', value: colorValue2 }, bold: true }
            },
            lastCol: {
                cell: { fillColor: { type: 'scheme', value: colorValue1 } },
                character: { color: { type: 'scheme', value: colorValue2 }, bold: true }
            },
            lastRow: {
                cell: { fillColor: { type: 'scheme', value: colorValue2 }, borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue3 } } }
            }
            // southEastCell:{
            //     character: { color: { type: 'scheme', value: colorValue3 } }
            // },
            // southWestCell:{
            //     character: { color: { type: 'scheme', value: colorValue3 } }
            // }
        };
    };

    // Medium Style 4 table style attributes (Presentation)
    TableStyles.getMediumStyle4TableStyleAttributesPresentation = function (colorValue1, colorValue2) {
        return {
            wholeTable: {
                cell: {
                    fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 20000 }] },
                    borderLeft: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderRight: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderInsideHor: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } },
                    borderInsideVert: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } }
                },
                character: { color: { type: 'scheme', value: colorValue2 } }
            },
            band1Hor: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 40000 }] } } },
            band1Vert: { cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 40000 }] } } },
            firstRow: {
                cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 20000 }] } },
                character: { bold: true }
            },
            lastRow: {
                cell: { fillColor: { type: 'scheme', value: colorValue1, transformations: [{ type: 'tint', value: 20000 }] }, borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue1 } } },
                character: { bold: true }
            }
        };
    };

    // Table grid, no style
    TableStyles.getTableGridPresentation = function (colorValue) {
        return {
            wholeTable: {
                cell: {
                    borderLeft: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } },
                    borderRight: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } },
                    borderBottom: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } },
                    borderTop: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } },
                    borderInsideHor: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } },
                    borderInsideVert: { style: 'single', width: 35, color: { type: 'scheme', value: colorValue } }
                }
            }
        };
    };

    // No grid, no style
    TableStyles.getTableNoGridNoStylePresentation = function (colorValue) {
        return {
            wholeTable: {
                cell: {
                    borderLeft: { style: 'none' },
                    borderRight: { style: 'none' },
                    borderBottom: { style: 'none' },
                    borderTop: { style: 'none' },
                    borderInsideHor: { style: 'none' },
                    borderInsideVert: { style: 'none' }
                },
                character: { fontName: 'Calibri Light', color: { type: 'scheme', value: colorValue }, bold: false }
            }
        };
    };

    // exports ================================================================

    // derive this class from class StyleCollection
    return StyleCollection.extend({ constructor: TableStyles });

});
